@@ -3,6 +3,7 @@ import json
 import logging.handlers
 import os
 import re
+from contextlib import contextmanager
 from copy import copy
 from dataclasses import asdict, dataclass, field, is_dataclass
 from enum import Enum
@@ -40,15 +41,37 @@ def _to_optional_pattern(regex: Optional[Union[str, Pattern]]) -> Optional[Patte
 
 class _SummariesJSONEncoder(json.JSONEncoder):
     def default(self, o):
+        if isinstance(o, TransactionSummary):
+            # special case for TransactionSummary, since the drivers are not serializable and dataclasses.asdict
+            # calls deepcopy
+            result = copy(o.__dict__)
+            _replace_driver_by_name(result)
+            return result
         if is_dataclass(o):
             return asdict(o)
         if isinstance(o, RemoteObjectProtocol):
-            return o.__dict__
+            result = copy(o.__dict__)
+            _replace_driver_by_name(result)
+            return result
         if isinstance(o, SyncObject):
-            return o.to_dict()
+            return o.to_dict(make_serializable=True)
         return str(o)
 
 
+def _replace_driver_by_name(obj):
+    # The driver object from libcloud stores a connection and is not serializable.
+    # Since sometimes we want to be able to deepcopy these things around,
+    # we replace the driver by its name. This is needed for `asdict` to work.
+    if isinstance(obj, RemoteObjectProtocol) and hasattr(obj, "driver"):
+        obj.driver = obj.driver.name
+    if isinstance(obj, list) or isinstance(obj, tuple):
+        for item in obj:
+            _replace_driver_by_name(item)
+    if isinstance(obj, dict):
+        for key, value in obj.items():
+            _replace_driver_by_name(value)
+
+
 class _JsonReprMixin:
     def to_json(self) -> str:
         return json.dumps(self, indent=2, sort_keys=True, cls=_SummariesJSONEncoder)
@@ -57,6 +80,19 @@ class _JsonReprMixin:
         return f"\n{self.__class__.__name__}: \n{self.to_json()}\n"
 
 
+@contextmanager
+def _switch_to_dir(path: str = None) -> bool:
+    if path:
+        cur_dir = os.getcwd()
+        try:
+            os.chdir(path)
+            yield
+        finally:
+            os.chdir(cur_dir)
+    else:
+        yield
+
+
 class Provider(str, Enum):
     GOOGLE_STORAGE = "google_storage"
     S3 = "s3"
@@ -96,6 +132,12 @@ class SyncObject(_JsonReprMixin):
         remote_obj: RemoteObjectProtocol = None,
         remote_path: str = None,
     ):
+        if remote_path is not None:
+            remote_path = remote_path.lstrip("/")
+        if remote_obj is not None:
+            remote_obj = copy(remote_obj)
+            remote_obj.name = remote_obj.name.lstrip("/")
+
         self.exists_locally = False
         self.local_path = None
         self.set_local_path(local_path)
@@ -129,9 +171,7 @@ class SyncObject(_JsonReprMixin):
     @property
     def exists_on_target(self) -> bool:
         """
-        Getter of the exists_on_target property. Since the file exists at least once (locally or remotely), the property
-        is True iff the file exists on both locations
-        :return: True iff the file exists on both locations
+        True iff the file exists on both locations
         """
         return self.exists_on_remote and self.exists_locally
 
@@ -152,24 +192,19 @@ class SyncObject(_JsonReprMixin):
 
     @property
     def exists_on_remote(self):
-        """
-        Getter of the exists_on_remote property. Is true if the file was found on the remote.
-        :return: whether the file exists on the remote
-        """
         return self.remote_obj is not None
 
     @property
     def equal_md5_hash_sum(self):
-        """
-        Getter of the equal_md5_hash_sum property.
-        :return: True if the local and the remote file have the same md5sum
-        """
         if self.exists_on_target:
             return self.local_hash == self.remote_obj.hash
         return False
 
-    def to_dict(self):
+    def to_dict(self, make_serializable=True):
         result = copy(self.__dict__)
+        if make_serializable:
+            _replace_driver_by_name(result)
+
         result["exists_on_remote"] = self.exists_on_remote
         result["exists_on_target"] = self.exists_on_target
         result["equal_md5_hash_sum"] = self.equal_md5_hash_sum
@@ -535,6 +570,8 @@ class RemoteStorage:
         if full_remote_path.endswith("/") or full_remote_path == "":
             return False
 
+        # Remove leading / for comparison of paths
+        full_remote_path = full_remote_path.lstrip("/")
         object_remote_path = RemoteStorage._get_remote_path(remote_object)
         is_in_selected_dir = object_remote_path.startswith(full_remote_path + "/")
         is_selected_file = object_remote_path == full_remote_path
@@ -685,10 +722,12 @@ class RemoteStorage:
             self.bucket.list_objects(full_remote_path)
         )
 
-        for obj in tqdm(remote_objects, desc="Remote paths: "):
+        for obj in tqdm(
+            remote_objects,
+            desc=f"Scanning remote paths in {self.bucket.name}/{full_remote_path}: ",
+        ):
             local_path = None
             collides_with = None
-            skip = False
             if (obj.size == 0) or (
                 self._listed_due_to_name_collision(full_remote_path, obj)
             ):
@@ -698,18 +737,9 @@ class RemoteStorage:
                 skip = True
             else:
                 relative_obj_path = self._get_relative_remote_path(obj)
-                if include_regex is not None:
-                    if not include_regex.match(relative_obj_path):
-                        log.debug(
-                            f"Skipping {relative_obj_path} due to regex {include_regex}"
-                        )
-                        skip = True
-                if exclude_regex is not None:
-                    if exclude_regex.match(relative_obj_path):
-                        log.debug(
-                            f"Skipping {relative_obj_path} due to regex {exclude_regex}"
-                        )
-                        skip = True
+                skip = self._should_skip(
+                    relative_obj_path, include_regex, exclude_regex
+                )
 
             if not skip:
                 local_path = self._get_destination_path(obj, local_base_dir)
@@ -722,40 +752,6 @@ class RemoteStorage:
 
         return summary
 
-    @staticmethod
-    def _get_push_local_path(path: str, local_path_prefix: Optional[str] = None) -> str:
-        """
-        Get the full local path of a file for pushing, including an optional path prefix.
-        Note that ``path`` may not be absolute if ``local_path_prefix`` is specified.
-
-        **Usage Examples:**
-
-        >>> RemoteStorage._get_push_local_path('/foo/bar/baz.txt')
-        '/foo/bar/baz.txt'
-        >>> RemoteStorage._get_push_local_path('foo/bar/baz.txt')
-        'foo/bar/baz.txt'
-        >>> RemoteStorage._get_push_local_path('bar/baz.txt', local_path_prefix='/foo')
-        '/foo/bar/baz.txt'
-        >>> RemoteStorage._get_push_local_path('/bar/baz.txt', local_path_prefix='/foo')
-        Traceback (most recent call last):
-        ...
-        ValueError: /bar/baz.txt is an absolute path and local_path_prefix was specified
-
-        :param path:
-        :param local_path_prefix:
-        :return: the full local path of the file
-        """
-        # Parameter validation
-        if local_path_prefix and Path(path).is_absolute():
-            raise ValueError(
-                f"{path} is an absolute path and local_path_prefix was specified"
-            )
-
-        if Path(path).is_absolute():
-            return path
-        else:
-            return os.path.join(local_path_prefix or "", path)
-
     def get_push_remote_path(self, local_path: str) -> str:
         """
         Get the full path within a remote storage bucket for pushing.
@@ -779,72 +775,97 @@ class RemoteStorage:
 
         :param path: Path to the local object (file or directory) to be uploaded, may be absolute or relative.
             globs are permitted, thus ``path`` may contain wildcards.
-        :param local_path_prefix: Prefix to be concatenated with ``path``
+        :param local_path_prefix: path names on the remote will be relative to this path. Thus, specifying
+            for example ``local_path_prefix=/bar/foo`` (on a unix system) and ``path=baz``
+            will push ``/bar/foo/baz`` to ``remote_base_path/baz``. The same will happen if
+            ``path=/bar/foo/baz`` is specified.
+            **NOTE**: if ``local_path_prefix`` is specified and ``path`` is absolute, it is assumed that
+            ``path`` is child of ``local_path_prefix``. If this is not the case, an error will be raised.
         :param include_regex: If not None, only files with paths matching the regex will be pushed.
+            Note that paths matched against the regex will be relative to ``local_path_prefix``.
         :param exclude_regex: If not None, only files with paths not matching the regex will be pushed.
             Takes precedence over ``include_regex``, i.e. if a file matches both regexes, it will be excluded.
+            Note that paths matched against the regex will be relative to ``local_path_prefix``.
         :param path_regex: DEPRECATED! Same as ``include_regex``.
         :return: the summary object
         """
         summary = TransactionSummary(sync_direction="push")
         include_regex = self._handle_deprecated_path_regex(include_regex, path_regex)
 
+        if local_path_prefix is not None:
+            local_path_prefix = os.path.abspath(local_path_prefix)
         include_regex = _to_optional_pattern(include_regex)
         exclude_regex = _to_optional_pattern(exclude_regex)
 
-        # collect all paths to scan
-        all_files_analyzed = []
-        for local_path in glob.glob(path):
-            local_path = self._get_push_local_path(local_path, local_path_prefix)
-            if os.path.isfile(local_path):
-                all_files_analyzed.append(local_path)
-            elif os.path.isdir(local_path):
-                for root, _, fs in os.walk(local_path):
-                    all_files_analyzed.extend([os.path.join(root, f) for f in fs])
-        if len(all_files_analyzed) == 0:
-            raise FileNotFoundError(
-                f"No files found under {path=} with {local_path_prefix=}"
-            )
-
-        for file in tqdm(all_files_analyzed, desc="Scanning file: "):
-            skip = False
-            collides_with = None
-            remote_obj = None
-            if include_regex is not None and not include_regex.match(file):
-                log.debug(
-                    f"Skipping {file} since it does not match regular expression '{include_regex}'."
+        _path = Path(path)
+        if _path.is_absolute() and local_path_prefix:
+            try:
+                path = str(_path.relative_to(local_path_prefix))
+            except ValueError:
+                raise ValueError(
+                    f"Specified {path=} is not a child of {local_path_prefix=}"
                 )
-                skip = True
-            if exclude_regex is not None and exclude_regex.match(file):
-                log.debug(
-                    f"Skipping {file} since it matches regular expression '{exclude_regex}'."
+
+        # at this point, path is relative to local_path_prefix.
+        with _switch_to_dir(local_path_prefix):
+            # collect all paths to scan
+            all_files_analyzed = []
+            for local_path in glob.glob(path):
+                if os.path.isfile(local_path):
+                    all_files_analyzed.append(local_path)
+                elif os.path.isdir(local_path):
+                    for root, _, fs in os.walk(local_path):
+                        all_files_analyzed.extend([os.path.join(root, f) for f in fs])
+            if len(all_files_analyzed) == 0:
+                raise FileNotFoundError(
+                    f"No files found under {path=} with {local_path_prefix=}"
                 )
-                skip = True
 
-            remote_path = self.get_push_remote_path(file)
-            # noinspection PyTypeChecker
-            matched_remote_obj = [
-                obj
-                for obj in self.bucket.list_objects(remote_path)
-                if not self._listed_due_to_name_collision(remote_path, obj)
-            ]
+            for file in tqdm(
+                all_files_analyzed, desc=f"Scanning files in {os.getcwd()}: "
+            ):
+                collides_with = None
+                remote_obj = None
+                skip = self._should_skip(file, include_regex, exclude_regex)
 
-            # name collision of local file with remote dir
-            if len(matched_remote_obj) > 1:
-                collides_with = matched_remote_obj
+                remote_path = self.get_push_remote_path(file)
+                # noinspection PyTypeChecker
+                matched_remote_obj = [
+                    obj
+                    for obj in self.bucket.list_objects(remote_path)
+                    if not self._listed_due_to_name_collision(remote_path, obj)
+                ]
 
-            elif matched_remote_obj:
-                remote_obj = matched_remote_obj[0]
+                # name collision of local file with remote dir
+                if len(matched_remote_obj) > 1:
+                    collides_with = matched_remote_obj
 
-            synced_obj = SyncObject(file, remote_obj, remote_path=remote_path)
-            summary.add_entry(
-                synced_obj,
-                collides_with=collides_with,
-                skip=skip,
-            )
+                elif matched_remote_obj:
+                    remote_obj = matched_remote_obj[0]
+
+                synced_obj = SyncObject(file, remote_obj, remote_path=remote_path)
+                summary.add_entry(
+                    synced_obj,
+                    collides_with=collides_with,
+                    skip=skip,
+                )
 
         return summary
 
+    @staticmethod
+    def _should_skip(file: str, include_regex: Pattern, exclude_regex: Pattern):
+        if include_regex is not None and not include_regex.match(file):
+            log.debug(
+                f"Skipping {file} since it does not match regular expression '{include_regex}'."
+            )
+            return True
+        if exclude_regex is not None and exclude_regex.match(file):
+            log.debug(
+                f"Skipping {file} since it matches regular expression '{exclude_regex}'."
+            )
+            return True
+        return False
+
     @staticmethod
     def _handle_deprecated_path_regex(
         include_regex: Optional[Union[Pattern, str]],
@@ -874,10 +895,11 @@ class RemoteStorage:
         path_regex: Union[Pattern, str] = None,
     ) -> TransactionSummary:
         """
-        Upload a local file or directory into the remote storage.
+        Upload files into the remote storage.
         Does not upload files for which the md5sum matches existing remote files.
         The remote path for uploading will be constructed from the remote_base_path and the provided path.
-        The local_path_prefix serves for finding the directory on the local system.
+        The local_path_prefix serves for finding the directory on the local system or for stripping off
+        parts of absolute paths if path is absolute, see examples below.
 
         Examples:
            1) path=foo/bar, local_path_prefix=None -->
@@ -886,8 +908,10 @@ class RemoteStorage:
                 /home/foo/bar uploaded to remote_base_path/home/foo/bar
            3) path=bar, local_path_prefix=/home/foo -->
                 /home/foo/bar uploaded to remote_base_path/bar
-
-        Note that ``path`` may not be absolute if ``local_path_prefix`` is specified.
+           4) path=/home/foo/bar, local_path_prefix=/home/foo -->
+                /home/foo/bar uploaded to remote_base_path/bar (Same as 3)
+           5) path=/home/baz/bar, local_path_prefix=/home/foo -->
+                ValueError: Specified path=/home/baz/bar is not a child of local_path_prefix=/home/foo
 
         :param path: Path to the local object (file or directory) to be uploaded, may be absolute or relative.
             globs are supported as well, thus ``path`` may be a pattern like ``*.txt``.
@@ -895,10 +919,12 @@ class RemoteStorage:
         :param force: If False, push will raise an error if an already existing remote file deviates from the local
             in its md5sum. If True, these files are overwritten.
         :param include_regex: If not None, only files with paths matching the regex will be pushed.
+            Note that paths matched against the regex will be relative to ``local_path_prefix``.
         :param exclude_regex: If not None, only files with paths not matching the regex will be pushed. Takes precedence
             over ``include_regex``, i.e. if a file matches both regexes, it will be excluded.
+            Note that paths matched against the regex will be relative to ``local_path_prefix``.
         :param dryrun: If True, simulates the push operation and returns the summary
-            (with synced_files being an empty list). Same as get_push_summary method.
+            (with synced_files being an empty list).
         :param path_regex: DEPRECATED! Same as ``include_regex``.
         :return: An object describing the summary of the operation.
         """
@@ -919,7 +945,7 @@ class RemoteStorage:
         path_regex: Union[Pattern, str] = None,
     ) -> List[RemoteObjectProtocol]:
         """
-        Deletes a file or a directory under the given path relative to local_base_dir. Use with caution.
+        Deletes a file or a directory under the given path relative to local_base_dir. Use with caution!
 
         :param remote_path: remote path on storage bucket relative to the configured remote base path.
         :param include_regex: If not None only files with paths matching the regex will be deleted.
